import fugue
import fugue.api as fa
import numpy as np
import pandas as pd
import pytest
import time

from nixtla.nixtla_client import NixtlaClient
from typing import Callable

# setting used for distributed related tests
ATOL = 1e-3


# test num partitions
# we need to be sure that we can recover the same results
# using a for loop
# A: be aware that num partitons can produce different results
# when used finetune_steps
def check_num_partitions_same_results(method: Callable, num_partitions: int, **kwargs):
    res_partitioned = method(**kwargs, num_partitions=num_partitions)
    res_no_partitioned = method(**kwargs, num_partitions=1)
    sort_by = ["unique_id", "ds"]
    if "cutoff" in res_partitioned:
        sort_by.extend(["cutoff"])
    pd.testing.assert_frame_equal(
        res_partitioned.sort_values(sort_by).reset_index(drop=True),
        res_no_partitioned.sort_values(sort_by).reset_index(drop=True),
        rtol=1e-2,
        atol=1e-2,
    )


def check_retry_behavior(
    df,
    side_effect,
    side_effect_exception,
    max_retries=5,
    retry_interval=5,
    max_wait_time=40,
    should_retry=True,
    sleep_seconds=5,
):
    mock_nixtla_client = NixtlaClient(
        max_retries=max_retries,
        retry_interval=retry_interval,
        max_wait_time=max_wait_time,
    )
    mock_nixtla_client._make_request = side_effect
    init_time = time.time()
    with pytest.raises(side_effect_exception):
        mock_nixtla_client.forecast(
            df=df, h=12, time_col="timestamp", target_col="value"
        )
    total_mock_time = time.time() - init_time
    if should_retry:
        approx_expected_time = min((max_retries - 1) * retry_interval, max_wait_time)
        upper_expected_time = min(max_retries * retry_interval, max_wait_time)
        assert total_mock_time >= approx_expected_time, "It is not retrying as expected"
        # preprocessing time before the first api call should be less than 60 seconds
        assert (
            total_mock_time - upper_expected_time - (max_retries - 1) * sleep_seconds
            <= sleep_seconds
        )
    else:
        assert total_mock_time <= max_wait_time


# test we recover the same <mean> forecasts
# with and without restricting input
# (add_history)
def check_equal_fcsts_add_history(nixtla_client, **kwargs):
    fcst_no_rest_df = nixtla_client.forecast(**kwargs, add_history=True)
    fcst_no_rest_df = (
        fcst_no_rest_df.groupby("unique_id", observed=True)
        .tail(kwargs["h"])
        .reset_index(drop=True)
    )
    fcst_rest_df = nixtla_client.forecast(**kwargs)
    pd.testing.assert_frame_equal(
        fcst_no_rest_df,
        fcst_rest_df,
        atol=1e-4,
        rtol=1e-3,
    )
    return fcst_rest_df


def check_quantiles(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    id_col: str = "id_col",
    time_col: str = "time_col",
):
    test_qls = list(np.arange(0.1, 1, 0.1))
    exp_q_cols = [f"TimeGPT-q-{int(q * 100)}" for q in test_qls]

    def test_method_qls(method, **kwargs):
        df_qls = method(
            df=df, h=12, id_col=id_col, time_col=time_col, quantiles=test_qls, **kwargs
        )
        df_qls = fa.as_pandas(df_qls)
        assert all(col in df_qls.columns for col in exp_q_cols)
        # test monotonicity of quantiles
        df_qls.apply(lambda x: x.is_monotonic_increasing, axis=1).sum() == len(
            exp_q_cols
        )

    test_method_qls(nixtla_client.forecast)
    test_method_qls(nixtla_client.forecast, add_history=True)
    test_method_qls(nixtla_client.cross_validation)


def check_cv_same_results_num_partitions(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    horizon: int = 12,
    id_col: str = "unique_id",
    time_col: str = "ds",
    **fcst_kwargs,
):
    fcst_df = nixtla_client.cross_validation(
        df=df,
        h=horizon,
        num_partitions=1,
        id_col=id_col,
        time_col=time_col,
        **fcst_kwargs,
    )
    fcst_df = fa.as_pandas(fcst_df)
    fcst_df_2 = nixtla_client.cross_validation(
        df=df,
        h=horizon,
        num_partitions=2,
        id_col=id_col,
        time_col=time_col,
        **fcst_kwargs,
    )
    fcst_df_2 = fa.as_pandas(fcst_df_2)
    pd.testing.assert_frame_equal(
        fcst_df.sort_values([id_col, time_col]).reset_index(drop=True),
        fcst_df_2.sort_values([id_col, time_col]).reset_index(drop=True),
        atol=ATOL,
    )


def check_forecast_diff_results_diff_models(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    horizon: int = 12,
    id_col: str = "unique_id",
    time_col: str = "ds",
    **fcst_kwargs,
):
    fcst_df = nixtla_client.forecast(
        df=df,
        h=horizon,
        num_partitions=1,
        id_col=id_col,
        time_col=time_col,
        model="timegpt-1",
        **fcst_kwargs,
    )
    fcst_df = fa.as_pandas(fcst_df)
    fcst_df_2 = nixtla_client.forecast(
        df=df,
        h=horizon,
        num_partitions=1,
        id_col=id_col,
        time_col=time_col,
        model="timegpt-1-long-horizon",
        **fcst_kwargs,
    )
    fcst_df_2 = fa.as_pandas(fcst_df_2)

    with pytest.raises(
        AssertionError, match=r'\(column name="TimeGPT"\) are different'
    ):
        pd.testing.assert_frame_equal(
            fcst_df.sort_values([id_col, time_col]).reset_index(drop=True),
            fcst_df_2.sort_values([id_col, time_col]).reset_index(drop=True),
        )


def check_forecast(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    horizon: int = 12,
    id_col: str = "unique_id",
    time_col: str = "ds",
    n_series_to_check: int = 4,
    **fcst_kwargs,
):
    fcst_df = nixtla_client.forecast(
        df=df,
        h=horizon,
        id_col=id_col,
        time_col=time_col,
        **fcst_kwargs,
    )
    fcst_df = fa.as_pandas(fcst_df)
    assert n_series_to_check * 12 == len(fcst_df)
    cols = fcst_df.columns.to_list()
    exp_cols = [id_col, time_col, "TimeGPT"]
    if "level" in fcst_kwargs:
        level = sorted(fcst_kwargs["level"])
        exp_cols.extend([f"TimeGPT-lo-{lv}" for lv in reversed(level)])
        exp_cols.extend([f"TimeGPT-hi-{lv}" for lv in level])
    assert cols == exp_cols


def check_forecast_same_results_num_partitions(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    horizon: int = 12,
    id_col: str = "unique_id",
    time_col: str = "ds",
    **fcst_kwargs,
):
    fcst_df = nixtla_client.forecast(
        df=df,
        h=horizon,
        num_partitions=1,
        id_col=id_col,
        time_col=time_col,
        **fcst_kwargs,
    )
    fcst_df = fa.as_pandas(fcst_df)
    fcst_df_2 = nixtla_client.forecast(
        df=df,
        h=horizon,
        num_partitions=2,
        id_col=id_col,
        time_col=time_col,
        **fcst_kwargs,
    )
    fcst_df_2 = fa.as_pandas(fcst_df_2)
    pd.testing.assert_frame_equal(
        fcst_df.sort_values([id_col, time_col]).reset_index(drop=True),
        fcst_df_2.sort_values([id_col, time_col]).reset_index(drop=True),
        atol=ATOL,
    )


def check_forecast_dataframe(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    n_series_to_check: int = 4,
):
    check_cv_same_results_num_partitions(nixtla_client, df, n_windows=2, step_size=1)
    check_cv_same_results_num_partitions(
        nixtla_client, df, n_windows=3, step_size=None, horizon=1
    )
    check_cv_same_results_num_partitions(
        nixtla_client, df, model="timegpt-1-long-horizon", horizon=1
    )
    check_forecast_diff_results_diff_models(nixtla_client, df)
    check_forecast(nixtla_client, df, num_partitions=1)
    check_forecast(
        nixtla_client,
        df,
        level=[90, 80],
        num_partitions=1,
        n_series_to_check=n_series_to_check,
    )
    check_forecast_same_results_num_partitions(nixtla_client, df)


def check_forecast_dataframe_diff_cols(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    id_col: str = "id_col",
    time_col: str = "time_col",
    target_col: str = "target_col",
):
    check_forecast(
        nixtla_client,
        df,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        num_partitions=1,
    )
    check_forecast(
        nixtla_client,
        df,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        level=[90, 80],
        num_partitions=1,
    )
    check_forecast_same_results_num_partitions(
        nixtla_client, df, id_col=id_col, time_col=time_col, target_col=target_col
    )


def check_anomalies(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    id_col: str = "unique_id",
    time_col: str = "ds",
    target_col: str = "y",
    **anomalies_kwargs,
):
    anomalies_df = nixtla_client.detect_anomalies(
        df=df,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        **anomalies_kwargs,
    )
    anomalies_df = fa.as_pandas(anomalies_df)
    assert (fa.as_pandas(df)[id_col].unique() == anomalies_df[id_col].unique()).all()
    cols = anomalies_df.columns.to_list()
    level = anomalies_kwargs.get("level", 99)
    exp_cols = [
        id_col,
        time_col,
        target_col,
        "TimeGPT",
        "anomaly",
        f"TimeGPT-lo-{level}",
        f"TimeGPT-hi-{level}",
    ]
    assert cols == exp_cols


def check_anomalies_same_results_num_partitions(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    id_col: str = "unique_id",
    time_col: str = "ds",
    target_col: str = "y",
    **anomalies_kwargs,
):
    anomalies_df = nixtla_client.detect_anomalies(
        df=df,
        num_partitions=1,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        **anomalies_kwargs,
    )
    anomalies_df = fa.as_pandas(anomalies_df)
    anomalies_df_2 = nixtla_client.detect_anomalies(
        df=df,
        num_partitions=2,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        **anomalies_kwargs,
    )
    anomalies_df_2 = fa.as_pandas(anomalies_df_2)
    pd.testing.assert_frame_equal(
        anomalies_df.sort_values([id_col, time_col]).reset_index(drop=True),
        anomalies_df_2.sort_values([id_col, time_col]).reset_index(drop=True),
        atol=ATOL,
    )


def check_anomalies_dataframe(nixtla_client: NixtlaClient, df: fugue.AnyDataFrame):
    check_anomalies(nixtla_client, df, num_partitions=1)
    check_anomalies(nixtla_client, df, level=90, num_partitions=1)
    check_anomalies_same_results_num_partitions(nixtla_client, df)


def check_online_anomalies(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    id_col: str = "unique_id",
    time_col: str = "ds",
    target_col: str = "y",
    level=99,
    **reatlime_anomalies_kwargs,
):
    anomalies_df = nixtla_client.detect_anomalies_online(
        df=df,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        **reatlime_anomalies_kwargs,
    )
    anomalies_df = fa.as_pandas(anomalies_df)
    assert (fa.as_pandas(df)[id_col].unique() == anomalies_df[id_col].unique()).all()
    cols = anomalies_df.columns.to_list()
    exp_cols = [
        id_col,
        time_col,
        target_col,
        "TimeGPT",
        "anomaly",
        "anomaly_score",
        f"TimeGPT-lo-{level}",
        f"TimeGPT-hi-{level}",
    ]
    assert cols == exp_cols


def check_anomalies_online_same_results_num_partitions(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    id_col: str = "unique_id",
    time_col: str = "ds",
    target_col: str = "y",
    **reatlime_anomalies_kwargs,
):
    anomalies_df = nixtla_client.detect_anomalies_online(
        df=df,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        num_partitions=1,
        **reatlime_anomalies_kwargs,
    )
    anomalies_df = fa.as_pandas(anomalies_df)
    anomalies_df_2 = nixtla_client.detect_anomalies_online(
        df=df,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        num_partitions=2,
        **reatlime_anomalies_kwargs,
    )
    anomalies_df_2 = fa.as_pandas(anomalies_df_2)
    pd.testing.assert_frame_equal(
        anomalies_df.sort_values([id_col, time_col]).reset_index(drop=True),
        anomalies_df_2.sort_values([id_col, time_col]).reset_index(drop=True),
        atol=ATOL,
    )


def check_anomalies_online_dataframe(
    nixtla_client: NixtlaClient, df: fugue.AnyDataFrame
):
    check_online_anomalies(
        nixtla_client,
        df,
        h=20,
        detection_size=5,
        threshold_method="univariate",
        level=99,
        num_partitions=1,
    )
    check_anomalies_online_same_results_num_partitions(
        nixtla_client,
        df,
        h=20,
        detection_size=5,
        threshold_method="univariate",
        level=99,
    )


def check_anomalies_dataframe_diff_cols(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    id_col: str = "id_col",
    time_col: str = "time_col",
    target_col: str = "target_col",
):
    check_anomalies(
        nixtla_client,
        df,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        num_partitions=1,
    )
    check_anomalies(
        nixtla_client,
        df,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        level=90,
        num_partitions=1,
    )
    check_anomalies_same_results_num_partitions(
        nixtla_client, df, id_col=id_col, time_col=time_col, target_col=target_col
    )


def check_forecast_x(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    X_df: fugue.AnyDataFrame,
    horizon: int = 24,
    id_col: str = "unique_id",
    time_col: str = "ds",
    target_col: str = "y",
    **fcst_kwargs,
):
    fcst_df = nixtla_client.forecast(
        df=df,
        X_df=X_df,
        h=horizon,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        **fcst_kwargs,
    )
    fcst_df = fa.as_pandas(fcst_df)
    n_series = fa.as_pandas(X_df)[id_col].nunique()
    assert n_series * horizon == len(fcst_df)

    cols = fcst_df.columns.to_list()
    exp_cols = [id_col, time_col, "TimeGPT"]
    if "level" in fcst_kwargs:
        level = sorted(fcst_kwargs["level"])
        exp_cols.extend([f"TimeGPT-lo-{lv}" for lv in reversed(level)])
        exp_cols.extend([f"TimeGPT-hi-{lv}" for lv in level])
    assert cols == exp_cols

    fcst_df_2 = nixtla_client.forecast(
        df=df,
        h=horizon,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        **fcst_kwargs,
    )
    fcst_df_2 = fa.as_pandas(fcst_df_2)
    equal_arrays = np.array_equal(
        fcst_df.sort_values([id_col, time_col])["TimeGPT"].values,
        fcst_df_2.sort_values([id_col, time_col])["TimeGPT"].values,
    )
    assert not equal_arrays, "Forecasts with and without ex vars are equal"


def check_forecast_x_same_results_num_partitions(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    X_df: fugue.AnyDataFrame,
    horizon: int = 24,
    id_col: str = "unique_id",
    time_col: str = "ds",
    target_col: str = "y",
    **fcst_kwargs,
):
    fcst_df = nixtla_client.forecast(
        df=df,
        X_df=X_df,
        h=horizon,
        num_partitions=1,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        **fcst_kwargs,
    )
    fcst_df = fa.as_pandas(fcst_df)
    fcst_df_2 = nixtla_client.forecast(
        df=df,
        h=horizon,
        num_partitions=2,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        **fcst_kwargs,
    )
    fcst_df_2 = fa.as_pandas(fcst_df_2)
    equal_arrays = np.array_equal(
        fcst_df.sort_values([id_col, time_col])["TimeGPT"].values,
        fcst_df_2.sort_values([id_col, time_col])["TimeGPT"].values,
    )
    assert not equal_arrays, "Forecasts with and without ex vars are equal"


def check_forecast_x_dataframe(
    nixtla_client: NixtlaClient, df: fugue.AnyDataFrame, X_df: fugue.AnyDataFrame
):
    check_forecast_x(nixtla_client, df, X_df, num_partitions=1)
    check_forecast_x(nixtla_client, df, X_df, level=[90, 80], num_partitions=1)
    check_forecast_x_same_results_num_partitions(nixtla_client, df, X_df)


def check_forecast_x_dataframe_diff_cols(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    X_df: fugue.AnyDataFrame,
    id_col: str = "id_col",
    time_col: str = "time_col",
    target_col: str = "target_col",
):
    check_forecast_x(
        nixtla_client,
        df,
        X_df,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        num_partitions=1,
    )
    check_forecast_x(
        nixtla_client,
        df,
        X_df,
        id_col=id_col,
        time_col=time_col,
        target_col=target_col,
        level=[90, 80],
        num_partitions=1,
    )
    check_forecast_x_same_results_num_partitions(
        nixtla_client, df, X_df, id_col=id_col, time_col=time_col, target_col=target_col
    )


def check_finetuned_model(
    nixtla_client: NixtlaClient,
    df: fugue.AnyDataFrame,
    model_id2: str,
):
    # fine-tuning on distributed fails
    with pytest.raises(
        ValueError, match="Can only fine-tune on pandas or polars dataframes."
    ):
        nixtla_client.finetune(df=df)

    # forecast
    local_fcst = nixtla_client.forecast(
        df=fa.as_pandas(df), h=5, finetuned_model_id=model_id2
    )
    distr_fcst = (
        fa.as_pandas(nixtla_client.forecast(df=df, h=5, finetuned_model_id=model_id2))
        .sort_values(["unique_id", "ds"])
        .reset_index(drop=True)
    )
    pd.testing.assert_frame_equal(
        local_fcst,
        distr_fcst,
        check_dtype=False,
        atol=1e-4,
        rtol=1e-2,
    )

    # cross-validation
    local_cv = nixtla_client.cross_validation(
        df=fa.as_pandas(df), n_windows=2, h=5, finetuned_model_id=model_id2
    )
    distr_cv = (
        fa.as_pandas(
            nixtla_client.cross_validation(
                df=df, n_windows=2, h=5, finetuned_model_id=model_id2
            )
        )
        .sort_values(["unique_id", "ds"])
        .reset_index(drop=True)
    )
    pd.testing.assert_frame_equal(
        local_cv,
        distr_cv[local_cv.columns],
        check_dtype=False,
        atol=1e-4,
        rtol=1e-2,
    )

    # anomaly detection
    local_anomaly = nixtla_client.detect_anomalies(
        df=fa.as_pandas(df), finetuned_model_id=model_id2
    )
    distr_anomaly = (
        fa.as_pandas(
            nixtla_client.detect_anomalies(df=df, finetuned_model_id=model_id2)
        )
        .sort_values(["unique_id", "ds"])
        .reset_index(drop=True)
    )
    pd.testing.assert_frame_equal(
        local_anomaly,
        distr_anomaly[local_anomaly.columns],
        check_dtype=False,
        atol=1e-3,
        rtol=1e-2,
    )



import warnings
from dataclasses import dataclass, asdict
from functools import partial
from pathlib import Path
from typing import Any, Callable, List

import pandas as pd
from utilsforecast.evaluation import evaluate
from utilsforecast.losses import mae, _zero_to_nan

from .logger_config import setup_logger

warnings.simplefilter(
    action="ignore",
    category=FutureWarning,
)
main_logger = setup_logger(__name__)


def mase(
    df: pd.DataFrame,
    models: List[str],
    seasonality: int,
    train_df: pd.DataFrame,
    id_col: str = "unique_id",
    target_col: str = "y",
) -> pd.DataFrame:
    mean_abs_err = mae(df, models, id_col, target_col)
    mean_abs_err = mean_abs_err.set_index(id_col)
    # assume train_df is sorted
    lagged = train_df.groupby(id_col, observed=True)[target_col].shift(seasonality)
    scale = train_df[target_col].sub(lagged).abs()
    scale = scale.groupby(train_df[id_col], observed=True).mean()
    scale[scale < 1e-2] = 0.0
    res = mean_abs_err.div(_zero_to_nan(scale), axis=0).fillna(0)
    res.index.name = id_col
    res = res.reset_index()
    return res


def generate_train_cv_splits(
    df: pd.DataFrame,
    cutoffs: pd.DataFrame,
) -> pd.DataFrame:
    """
    based on `cutoffs` (columns `unique_id`, `cutoffs`)
    generates train cv splits using `df`
    """
    df = df.merge(cutoffs, on="unique_id", how="outer")
    df = df.query("ds <= cutoff")
    df = df.reset_index(drop=True)
    return df


@dataclass
class DatasetParams:
    frequency: str
    pandas_frequency: str
    horizon: int
    seasonality: int

    @staticmethod
    def _get_value_from_df_col(
        df: pd.DataFrame,
        col: str,
        dtype: Callable | None = None,
    ) -> Any:
        col_values = df[col].unique()
        if len(col_values) > 1:
            raise ValueError(f"{col} is not unique: {col_values}")
        value = col_values[0]
        if dtype is not None:
            value = dtype(value)
        return value

    @classmethod
    def from_df(cls, df: pd.DataFrame) -> "DatasetParams":
        dataset_params = {}
        dataset_params_cols = [
            "frequency",
            "pandas_frequency",
            "horizon",
            "seasonality",
        ]
        dataset_params_cols_dtypes = [str, str, int, int]
        for col, dtype in zip(dataset_params_cols, dataset_params_cols_dtypes):
            dataset_params[col] = cls._get_value_from_df_col(df, col, dtype=dtype)
        return cls(**dataset_params)


@dataclass
class ExperimentDataset(DatasetParams):
    df: pd.DataFrame

    @classmethod
    def from_df(cls, df: pd.DataFrame) -> "ExperimentDataset":
        """
        Parameters
        ----------
        df : pd.DataFrame
            df should have columns:
            unique_id, ds, y, frequency, pandas_frequency, horizon, seasonality
        """
        ds_params = DatasetParams.from_df(df=df)
        df = df[["unique_id", "ds", "y"]]  # type: ignore
        return cls(
            df=df,
            **asdict(ds_params),
        )

    @classmethod
    def from_parquet(
        cls,
        parquet_path: str | Path,
    ) -> "ExperimentDataset":
        df = pd.read_parquet(parquet_path)
        return cls.from_df(df=df)

    def evaluate_forecast_df(
        self,
        forecast_df: pd.DataFrame,
        models: List[str],
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        forecast_df : pd.DataFrame
            df should have columns: unique_id, ds, cutoff, y, and models
        """
        for model in models:
            if forecast_df[model].isna().sum() > 0:
                print(forecast_df.loc[forecast_df[model].isna()]["unique_id"].unique())
                raise ValueError(f"model {model} has NaN values")
        cutoffs = forecast_df[["unique_id", "cutoff"]].drop_duplicates()
        train_cv_splits = generate_train_cv_splits(df=self.df, cutoffs=cutoffs)

        def add_id_cutoff(df: pd.DataFrame):
            df["id_cutoff"] = (
                df["unique_id"].astype(str) + "-" + df["cutoff"].astype(str)
            )

        for df in [cutoffs, train_cv_splits, forecast_df]:
            add_id_cutoff(df)
        partial_mase = partial(mase, seasonality=self.seasonality)
        eval_df = evaluate(
            df=forecast_df,
            train_df=train_cv_splits,
            metrics=[partial_mase],
            models=models,
            id_col="id_cutoff",
        )
        eval_df = eval_df.merge(cutoffs, on=["id_cutoff"])
        eval_df = eval_df.drop(columns=["id_cutoff"])
        eval_df = eval_df[["unique_id", "cutoff", "metric"] + models]
        return eval_df


@dataclass
class ForecastDataset:
    forecast_df: pd.DataFrame
    time_df: pd.DataFrame

    @classmethod
    def from_dir(cls, dir: str | Path):
        dir_ = Path(dir)
        forecast_df = pd.read_parquet(dir_ / "forecast_df.parquet")
        time_df = pd.read_parquet(dir_ / "time_df.parquet")
        return cls(forecast_df=forecast_df, time_df=time_df)

    @staticmethod
    def is_forecast_ready(dir: str | Path):
        dir_ = Path(dir)
        forecast_path = dir_ / "forecast_df.parquet"
        time_path = dir_ / "time_df.parquet"
        return forecast_path.exists() and time_path.exists()

    def save_to_dir(self, dir: str | Path):
        dir_ = Path(dir)
        dir_.mkdir(parents=True, exist_ok=True)
        self.forecast_df.to_parquet(dir_ / "forecast_df.parquet")
        self.time_df.to_parquet(dir_ / "time_df.parquet")



import pytest

from nixtla_tests.helpers.checks import check_anomalies_dataframe
from nixtla_tests.helpers.checks import check_anomalies_online_dataframe
from nixtla_tests.helpers.checks import check_anomalies_dataframe_diff_cols
from nixtla_tests.helpers.checks import check_forecast_dataframe
from nixtla_tests.helpers.checks import check_forecast_dataframe_diff_cols
from nixtla_tests.helpers.checks import check_forecast_x_dataframe
from nixtla_tests.helpers.checks import check_forecast_x_dataframe_diff_cols
from nixtla_tests.helpers.checks import check_quantiles

pytestmark = pytest.mark.distributed_run

def test_quantiles(nixtla_test_client, dask_df):
    check_quantiles(nixtla_test_client, dask_df, id_col="unique_id", time_col="ds")


def test_forecast(nixtla_test_client, dask_df, dask_diff_cols_df, distributed_n_series):
    check_forecast_dataframe(
        nixtla_test_client, dask_df, n_series_to_check=distributed_n_series
    )
    check_forecast_dataframe_diff_cols(nixtla_test_client, dask_diff_cols_df)


def test_anomalies(nixtla_test_client, dask_df, dask_diff_cols_df):
    check_anomalies_dataframe(nixtla_test_client, dask_df)
    check_anomalies_dataframe_diff_cols(nixtla_test_client, dask_diff_cols_df)


def test_anomalies_online(nixtla_test_client, dask_df):
    check_anomalies_online_dataframe(nixtla_test_client, dask_df)


def test_forecast_x_dataframe(
    nixtla_test_client,
    dask_df_x,
    dask_future_ex_vars_df,
    dask_df_x_diff_cols,
    dask_future_ex_vars_df_diff_cols,
):
    check_forecast_x_dataframe(nixtla_test_client, dask_df_x, dask_future_ex_vars_df)
    check_forecast_x_dataframe_diff_cols(
        nixtla_test_client,
        dask_df_x_diff_cols,
        dask_future_ex_vars_df_diff_cols,
    )



"""
this module takes Nixtla's benchmarking data 
and filters it to prevent azureml from crashing
in the following cases:
- too short series, see https://learn.microsoft.com/en-us/azure/machine-learning/concept-automl-forecasting-methods?view=azureml-api-2#data-length-requirements
"""
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable

import fire
import numpy as np
import pandas as pd

logging.basicConfig(level=logging.INFO)
main_logger = logging.getLogger(__name__)


@dataclass
class DatasetParams:
    frequency: str
    pandas_frequency: str
    horizon: int
    seasonality: int

    @staticmethod
    def _get_value_from_df_col(
        df: pd.DataFrame,
        col: str,
        dtype: Callable | None = None,
    ) -> Any:
        col_values = df[col].unique()
        if len(col_values) > 1:
            raise ValueError(f"{col} is not unique: {col_values}")
        value = col_values[0]
        if dtype is not None:
            value = dtype(value)
        return value

    @classmethod
    def from_df(cls, df: pd.DataFrame) -> "DatasetParams":
        dataset_params = {}
        dataset_params_cols = [
            "frequency",
            "pandas_frequency",
            "horizon",
            "seasonality",
        ]
        dataset_params_cols_dtypes = [str, str, int, int]
        for col, dtype in zip(dataset_params_cols, dataset_params_cols_dtypes):
            dataset_params[col] = cls._get_value_from_df_col(df, col, dtype=dtype)
        return cls(**dataset_params)


def filter_and_clean_dataset(
    dataset_path: str,
    max_series: int = 1_000,
    n_train_cv: int = 2,
    n_seasonalities: int = 5,
    max_insample_length: int = 3_000,
    random_seed: int = 420,
):
    main_logger.info(f"Processing dataset {dataset_path}")
    df = pd.read_parquet(dataset_path)
    df = df.drop_duplicates(["unique_id", "ds"])  # type: ignore
    df = df.sort_values(["unique_id", "ds"])
    ds_params = DatasetParams.from_df(df)
    min_train_size_per_series = (
        ds_params.horizon
        + 2 * ds_params.horizon
        + (n_train_cv - 1) * ds_params.horizon
        + 1
    )
    if ds_params.seasonality < 100:
        # if series has low seasonality
        # we add n_seasonalities to min_train_size_per_series
        # to keep the series long enough
        min_train_size_per_series += n_seasonalities * ds_params.seasonality
    uids = df["unique_id"].unique()  # type: ignore
    df = (
        df.groupby("unique_id")
        .filter(lambda x: len(x) >= min_train_size_per_series)
        .groupby("unique_id")  # type: ignore
        .tail(max_insample_length + ds_params.horizon)
        .reset_index(drop=True)
    )
    main_logger.info(
        f"Filtering out {len(uids) - len(df['unique_id'].unique())} series"
    )
    uids = df["unique_id"].unique()  # type: ignore
    if len(uids) > max_series:
        np.random.seed(random_seed)
        uids = np.random.choice(uids, max_series, replace=False)  # type: ignore
        df = df.query("unique_id in @uids")  # type: ignore
        main_logger.info(f"Filtering out {len(uids) - max_series} series")
    # finally we clean some strange dates
    mask = df["ds"].str.endswith(":01")  # type: ignore
    df.loc[mask, "ds"] = df.loc[mask, "ds"].str[:-3] + ":00"
    # save the dataset
    dataset_path = Path(dataset_path)  # type: ignore
    filtered_dataset_path = dataset_path.parent / "filtered_datasets" / dataset_path.name  # type: ignore
    filtered_dataset_path.parent.mkdir(exist_ok=True, parents=True)
    df.to_parquet(filtered_dataset_path)
    main_logger.info(f"Filtered dataset saved to {filtered_dataset_path}")


if __name__ == "__main__":
    fire.Fire(filter_and_clean_dataset)



import re
from pathlib import Path

comment_pat = re.compile(r"<!--.*?-->", re.DOTALL)
anchor_pat = re.compile(r"<a.*?>(.*?)</a>")
output_path = Path("docs/mintlify")

# process docs
for file in Path("docs").glob("*.md"):
    text = file.read_text()
    text = comment_pat.sub("", text)
    text = anchor_pat.sub("", text)
    module_name = file.name.split(".")[-2]
    output_file = output_path / (module_name + ".mdx")
    output_file.write_text(text)


readme_text = Path("README.md").read_text()
readme_text = readme_text
(output_path / "index.mdx").write_text(readme_text)



import pytest

try:
    import ray
    from ray.cluster_utils import Cluster

    @pytest.fixture(scope="module")
    def ray_cluster_setup():
        ray_cluster = Cluster(initialize_head=True, head_node_args={"num_cpus": 2})
        with ray.init(address=ray_cluster.address, ignore_reinit_error=True):
            # add mock node to simulate a cluster
            ray_cluster.add_node(num_cpus=2)
            yield

    @pytest.fixture(scope="module")
    def ray_df(distributed_series):
        return ray.data.from_pandas(distributed_series)

    @pytest.fixture(scope="module")
    def ray_diff_cols_df(distributed_series, renamer):
        return ray.data.from_pandas(distributed_series.rename(columns=renamer))

    @pytest.fixture(scope="module")
    def ray_df_x(distributed_df_x):
        return ray.data.from_pandas(distributed_df_x)

    @pytest.fixture(scope="module")
    def ray_future_ex_vars_df(distributed_future_ex_vars_df):
        return ray.data.from_pandas(distributed_future_ex_vars_df)

    @pytest.fixture(scope="module")
    def ray_df_x_diff_cols(distributed_df_x, renamer):
        return ray.data.from_pandas(distributed_df_x.rename(columns=renamer))

    @pytest.fixture(scope="module")
    def ray_future_ex_vars_df_diff_cols(distributed_future_ex_vars_df, renamer):
        return ray.data.from_pandas(
            distributed_future_ex_vars_df.rename(columns=renamer)
        )
except ImportError:
    # If Ray is not installed, we skip the fixtures
    pytest.skip("Ray is not installed, skipping Ray fixtures", allow_module_level=True)



import logging
from pathlib import Path
from typing import List
from unicodedata import numeric

import fire
import pandas as pd
from rich.console import Console
from rich.table import Table

from src.utils.data_handler import ExperimentDataset, ForecastDataset

logging.basicConfig(level=logging.INFO)
main_logger = logging.getLogger(__name__)


def print_df_rich(df: pd.DataFrame):
    console = Console()
    table = Table()
    for col in df.select_dtypes(include=["float"]).columns:
        df[col] = df[col].apply(lambda x: f"{x:.3f}")
    for col in df.columns:
        table.add_column(col)
    for row in df.itertuples(index=False):
        table.add_row(*row)
    console.print(table)


METHODS = {
    "azure_automl": "automl_prediction",
    "nixtla_timegpt": "TimeGPT",
    "statsforecast_sn": "SeasonalNaive",
}


def get_model_name(method: str) -> str:
    if method not in METHODS:
        raise ValueError(f"Invalid method: {method}")
    return METHODS[method]


def evaluate_experiments(
    datasets_paths: str,
    methods_to_evaluate: List[str] = list(METHODS.keys()),
    results_dir: str = "./results",
):
    datasets_paths_ = datasets_paths.split(",")
    eval_datasets_df: pd.DataFrame | None = None
    for dataset_path in datasets_paths_:
        experiment_name = dataset_path.split("/")[-1].split(".")[0]
        eval_method_df: pd.DataFrame | None = None
        dataset: None | ExperimentDataset = None
        for method in methods_to_evaluate:
            results_experiment_dir = Path(results_dir) / method / experiment_name
            if ForecastDataset.is_forecast_ready(results_experiment_dir):
                main_logger.info(
                    f"Evaluating experiment {experiment_name} and method {method}"
                )
                forecast_dataset = ForecastDataset.from_dir(results_experiment_dir)
                if dataset is None:
                    dataset = ExperimentDataset.from_parquet(parquet_path=dataset_path)
                eval_df = dataset.evaluate_forecast_df(
                    forecast_df=forecast_dataset.forecast_df,
                    model=get_model_name(method),
                    total_time=forecast_dataset.total_time,
                )
                if eval_method_df is None:
                    eval_method_df = eval_df
                else:
                    eval_method_df = pd.concat(
                        [eval_method_df, eval_df],
                        axis=1,
                    )  # type: ignore
            else:
                main_logger.info(
                    f"Skipping evaluation for experiment {experiment_name} and method {method}"
                    " because the forecasts are not ready yet"
                )
        if eval_method_df is not None:
            eval_method_df.reset_index(inplace=True)
            eval_method_df.insert(0, "dataset", experiment_name)
            if eval_datasets_df is None:
                eval_datasets_df = eval_method_df
            else:
                eval_datasets_df = pd.concat(
                    [eval_datasets_df, eval_method_df],
                    ignore_index=True,
                )  # type: ignore
    if eval_datasets_df is not None:
        azure_renamer = {"automl_prediction": "AzureAutoML"}
        if "azure_automl" in methods_to_evaluate:
            eval_datasets_df = eval_datasets_df.rename(columns=azure_renamer)
        eval_datasets_df.to_csv(Path(results_dir) / "eval_datasets.csv", index=False)
        eval_datasets_df["metric"] = (
            eval_datasets_df["metric"].str.upper().str.replace("TOTAL_", "")
        )
        # scale by SeasonalNaive
        if "SeasonalNaive" in eval_datasets_df.columns:
            time_mask = eval_datasets_df["metric"] == "TIME"
            for model in eval_datasets_df.columns.drop(["dataset", "metric"]):
                if model == "SeasonalNaive":
                    continue
                eval_datasets_df.loc[~time_mask, model] = (
                    eval_datasets_df.loc[~time_mask, model]
                    / eval_datasets_df.loc[~time_mask, "SeasonalNaive"]
                )
            eval_datasets_df = eval_datasets_df.drop(columns=["SeasonalNaive"])

        def pivot_df(df: pd.DataFrame, col: str) -> pd.DataFrame:
            return df.pivot(
                index="dataset",
                columns="metric",
                values=col,
            )

        result_list = []
        models = []
        for method in methods_to_evaluate:
            if method == "statsforecast_sn":
                continue
            if method == "azure_automl":
                col = "AzureAutoML"
            else:
                col = get_model_name(method)
            pivotted_df = pivot_df(eval_datasets_df, col)
            result_list.append(pivotted_df)
            models.append(col)
        result = pd.concat(result_list, axis=1, keys=models)
        result = result.swaplevel(axis=1).sort_index(axis=1)
        flattened_columns = ["_".join(col) for col in result.columns.values]
        result.columns = flattened_columns
        result = result.reset_index()
        print_df_rich(result)


if __name__ == "__main__":
    fire.Fire(evaluate_experiments)



from concurrent.futures import ThreadPoolExecutor
from copy import deepcopy
from time import time
from typing import Optional

import fire
import numpy as np
import pandas as pd
from prophet import Prophet as _Prophet
from utilsforecast.processing import (
    backtest_splits,
    drop_index_if_pandas,
    join,
    maybe_compute_sort_indices,
    take_rows,
    vertical_concat,
)

from src.tools import ExperimentHandler


class ParallelForecaster:
    def _process_group(self, func, df, **kwargs):
        uid = df["unique_id"].iloc[0]
        _df = df.drop("unique_id", axis=1)
        res_df = func(_df, **kwargs)
        res_df.insert(0, "unique_id", uid)
        return res_df

    def _apply_parallel(self, df_grouped, func, **kwargs):
        results = []
        with ThreadPoolExecutor(max_workers=None) as executor:
            futures = [
                executor.submit(self._process_group, func, df, **kwargs)
                for _, df in df_grouped
            ]
            for future in futures:
                results.append(future.result())
        return pd.concat(results)

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        X_df: Optional[pd.DataFrame] = None,
    ):
        df_grouped = df.groupby("unique_id")
        return self._apply_parallel(
            df_grouped,
            self._local_forecast,
            h=h,
        )

    def cross_validation(
        self,
        df: pd.DataFrame,
        h: int,
        n_windows: int = 1,
        step_size: Optional[int] = None,
        **kwargs,
    ):
        df_grouped = df.groupby("unique_id")
        kwargs = {"h": h, "n_windows": n_windows, "step_size": step_size, **kwargs}
        return self._apply_parallel(
            df_grouped,
            self._local_cross_validation,
            **kwargs,
        )


class Prophet(_Prophet, ParallelForecaster):
    def __init__(
        self,
        freq: str,
        alias: str = "Prophet",
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.freq = freq
        self.alias = alias

    def _local_forecast(
        self,
        df: pd.DataFrame,
        h: int,
        X_df: Optional[pd.DataFrame] = None,
    ) -> pd.DataFrame:
        model = deepcopy(self)
        model.fit(df=df)
        future_df = model.make_future_dataframe(
            periods=h, include_history=False, freq=self.freq
        )
        if X_df is not None:
            future_df = future_df.merge(X_df, how="left")
        np.random.seed(1000)
        fcst_df = model.predict(future_df)
        fcst_df = fcst_df.rename({"yhat": self.alias}, axis=1)
        fcst_df = fcst_df[["ds", self.alias]]
        return fcst_df

    def _local_cross_validation(
        self,
        df: pd.DataFrame,
        h: int,
        n_windows: int = 1,
        step_size: Optional[int] = None,
    ) -> pd.DataFrame:
        df = df.copy()
        df["ds"] = pd.to_datetime(df["ds"])
        df.insert(0, "unique_id", "ts_0")
        # mlforecast cv code
        results = []
        sort_idxs = maybe_compute_sort_indices(df, "unique_id", "ds")
        if sort_idxs is not None:
            df = take_rows(df, sort_idxs)
        splits = backtest_splits(
            df,
            n_windows=n_windows,
            h=h,
            id_col="unique_id",
            time_col="ds",
            freq=pd.tseries.frequencies.to_offset(self.freq),
            step_size=h if step_size is None else step_size,
        )
        for i_window, (cutoffs, train, valid) in enumerate(splits):
            if len(valid.columns) > 3:
                # if we have uid, ds, y + exogenous vars
                train_future = valid.drop(columns="y")
            else:
                train_future = None
            y_pred = self._local_forecast(
                df=train[["ds", "y"]],
                h=h,
                X_df=train_future,
            )
            y_pred.insert(0, "unique_id", "ts_0")
            y_pred = join(y_pred, cutoffs, on="unique_id", how="left")
            result = join(
                valid[["unique_id", "ds", "y"]],
                y_pred,
                on=["unique_id", "ds"],
            )
            if result.shape[0] < valid.shape[0]:
                raise ValueError(
                    "Cross validation result produced less results than expected. "
                    "Please verify that the frequency parameter (freq) matches your series' "
                    "and that there aren't any missing periods."
                )
            results.append(result)
        out = vertical_concat(results)
        out = drop_index_if_pandas(out)
        first_out_cols = ["unique_id", "ds", "cutoff", "y"]
        remaining_cols = [c for c in out.columns if c not in first_out_cols]
        fcst_cv_df = out[first_out_cols + remaining_cols]
        return fcst_cv_df.drop(columns="unique_id")


def evaluate_experiment(file: str):
    exp_handler = ExperimentHandler(file=file, method="prophet")
    Y_df, freq, pandas_freq, h, seasonality = exp_handler.read_data()
    model_name = "Prophet"
    print(model_name)
    prophet = Prophet(freq=pandas_freq)
    start = time()
    Y_hat_df = prophet.cross_validation(
        df=Y_df,
        h=h,
        n_windows=1,
    )
    total_time = time() - start
    print(total_time)
    # evaluation
    eval_df, total_time_df = exp_handler.evaluate_model(
        Y_hat_df=Y_hat_df,
        model_name=model_name,
        total_time=total_time,
    )
    exp_handler.save_results(
        freq=freq,
        eval_df=eval_df,
        total_time_df=total_time_df,
    )


if __name__ == "__main__":
    fire.Fire(evaluate_experiment)



import os
from typing import Optional, Tuple

import pandas as pd
from utilsforecast.evaluation import evaluate
from utilsforecast.losses import mae, rmse


class ExperimentHandler:
    def __init__(self, file: str, method: str):
        self.file = file
        self.method = method

    @staticmethod
    def get_parameter(parameter: str, df: pd.DataFrame):
        parameter = df[parameter].unique()
        if len(parameter) > 1:
            raise ValueError(f"{parameter} is not unique: {parameter}")
        return parameter[0]

    def read_data(
        self,
        max_insample_length: int = 3_000,
    ) -> Tuple[pd.DataFrame, str, str, int, int]:
        df = pd.read_parquet(self.file)
        Y_df = df[["unique_id", "ds", "y"]].drop_duplicates(["unique_id", "ds"])
        Y_df = Y_df.sort_values(["unique_id", "ds"])
        Y_df = Y_df.groupby("unique_id").tail(
            max_insample_length
        )  # take only last 3_000 rows
        Y_df["ds"] = Y_df["ds"].str.replace(":01$", ":00", regex=True)
        freq = self.get_parameter("frequency", df)
        pandas_freq = self.get_parameter("pandas_frequency", df)
        h = self.get_parameter("horizon", df)
        seasonality = self.get_parameter("seasonality", df)
        return Y_df, freq, pandas_freq, int(h), int(seasonality)

    def evaluate_model(
        self,
        Y_hat_df: pd.DataFrame,
        model_name: str,
        total_time: float,
    ):
        if "cutoff" in Y_hat_df.columns:
            Y_hat_df = Y_hat_df.drop(columns="cutoff")
        eval_df = evaluate(
            df=Y_hat_df,
            metrics=[rmse, mae],
        )
        total_time_df = pd.DataFrame({"model": [model_name], "time": [total_time]})
        return eval_df, total_time_df

    @staticmethod
    def summarize_df(df: pd.DataFrame):
        n_unique_ids = df["unique_id"].nunique()
        mean_y = df["y"].mean()
        std_y = df["y"].std()
        lengths = df.groupby("unique_id").size()
        min_length = lengths.min()
        max_length = lengths.max()
        n_obs = len(df)
        summary = {
            "n_series": n_unique_ids,
            "mean": mean_y,
            "std": std_y,
            "min_length": min_length,
            "max_length": max_length,
            "n_obs": n_obs,
        }
        summary_df = pd.DataFrame.from_dict(summary, orient="index")
        summary_df = summary_df.transpose()
        return summary_df

    def save_results(
        self,
        freq: str,
        eval_df: pd.DataFrame,
        total_time_df: pd.DataFrame,
        df: Optional[pd.DataFrame] = None,
    ):
        eval_df["frequency"] = freq
        eval_df = eval_df.melt(
            id_vars=["frequency", "metric", "unique_id"],
            var_name="model",
            value_name="value",
        )
        total_time_df["frequency"] = freq
        dir = self.file.split("/")[-1].replace(".parquet", "")
        dir = f"./data/results/{dir}"
        os.makedirs(dir, exist_ok=True)
        eval_df.to_parquet(
            f"{dir}/{self.method}_metrics.parquet",
            index=False,
        )
        total_time_df.to_parquet(
            f"{dir}/{self.method}_time.parquet",
            index=False,
        )
        if df is not None:
            summary_df = self.summarize_df(df)
            summary_df["frequency"] = freq
            print(summary_df)
            summary_df.to_parquet(
                f"{dir}/series_summary.parquet",
                index=False,
            )



from copy import deepcopy
from typing import List
from threadpoolctl import threadpool_limits

import pandas as pd
from prophet import Prophet

from ..utils.parallel_forecaster import ParallelForecaster
from ..utils.forecaster import Forecaster


class NixtlaProphet(Prophet, ParallelForecaster, Forecaster):
    def __init__(
        self,
        alias: str = "Prophet",
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.alias = alias

    def __local_forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
        quantiles: List[float] | None = None,
    ) -> pd.DataFrame:
        if quantiles is not None:
            raise NotImplementedError
        model = deepcopy(self)
        model.fit(df=df)
        future_df = model.make_future_dataframe(
            periods=h,
            include_history=False,
            freq=freq,
        )
        fcst_df = model.predict(future_df)
        fcst_df = fcst_df.rename({"yhat": self.alias}, axis=1)
        fcst_df = fcst_df[["ds", self.alias]]
        return fcst_df

    def _local_forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
        quantiles: List[float] | None = None,
    ) -> pd.DataFrame:
        with threadpool_limits(limits=1):
            return self.__local_forecast(
                df=df,
                h=h,
                freq=freq,
                quantiles=quantiles,
            )
